import datetime
import hashlib
import json
import uuid

from cryptography.hazmat.primitives import serialization

from Mine.Card import Card
from Mine.Transaction import Transaction

class Block():
    def __init__(self, blockId=None, previousHash="0", timestamp=None, height=0, nonce=0,
                 difficulty=0, card=None, transactions=[], data=None):
        if data:
            self.load_from_data(data)
        else:
            delay = 2
            self.blockId = blockId if blockId else uuid.uuid4()
            self.previousHash = previousHash
            self.timestamp = timestamp if timestamp else datetime.datetime.now(
                datetime.timezone.utc
                )
            self.height = height
            self.difficulty = difficulty
            self.nonce = nonce
            while True:
                try:
                    self.card = card if card else Card()
                except ReadTimeout as e:
                    time.sleep(delay := delay**2)
                    continue
                break
            self.transactions = transactions
        self.update()

    def update(self):
        self.blockHash = hashlib.sha256(str(self).encode('utf-8'))

    def mine(self, authorPrivateKey):
        self.transactions.append(
            Transaction(
                cardId = self.card.cardId,
                sender = authorPrivateKey.public_key(),
                receiver = authorPrivateKey.public_key(),
                authorPrivateKey = authorPrivateKey
                )
            )
        while int(self.blockHash.hexdigest(), 16) > 2**(256-self.difficulty):
            self.nonce += 1
            self.update()

    ###
    # Validate the internal block structure.
    # This method confirms that the correct data types have been used and that
    # values are basically valid.
    # Further validations against the existing chain must be done elsewhere.
    # This method calls the Card.validate() and Transaction.validate() methods to validate the
    # respective schemas.
    ###
    def validate(self):
        if not isinstance(self.blockId, uuid.UUID):
            raise TypeError(f"Block ID should be a UUID not {type(self.blockId).__name__}.")
        if not self.blockId.version == 4:
            raise self.Invalid(f"Block ID version should be 4 not {self.blockId.version}.")
        if not isinstance(self.previousHash, str):
            raise TypeError(
                f"Previous hash should be string not {type(self.previousHash).__name__}."
                )
        if not int(self.blockHash.hexdigest(), 16) <= 2**(256-self.difficulty):
            raise self.Invalid("Hash does not meet difficulty requirement.")
        if not isinstance(self.timestamp, datetime.datetime):
            raise TypeError(
                f"Timestamp should be a datetime not {type(self.timestamp).__name__}"
                )
        if not self.timestamp.tzinfo == datetime.timezone.utc:
            raise self.Invalid(
                f"Timestamp timezone should be in UTC not {self.timestamp.tzinfo}."
                )
        if not self.timestamp < datetime.datetime.now(datetime.timezone.utc):
            raise self.Invalid(f"Timestamp {self.timestamp} in the future.")
        if not isinstance(self.height, int):
            raise TypeError(f"Height should be integer not {type(self.height).__name__}.")
        if not self.height >= 0:
            raise self.Invalid(f"Height {self.height} less than 0.")
        if not isinstance(self.difficulty, int):
            raise TypeError(
                f"Difficulty should be an integer not {type(self.difficulty).__name__}."
                )
        if not self.difficulty >= 0:
            raise self.Invalid(f"Difficulty {self.difficulty} less than 0.")
        if not isinstance(self.nonce, int):
            raise TypeError(f"Nonce should be an integer not {type(self.nonce).__name__}.")
        if not self.nonce >= 0:
            raise self.Invalid(f"Nonce {self.nonce} less than 0.")
        self.card.validate()
        seenTransactions = []
        for transaction in self.transactions:
            if transaction in seenTransactions:
                raise self.Invalid(
                    f"Contains duplicate transaction {transaction.transactionId}."
                    )
            transaction.validate()
            seenTransactions.append(transaction)
        # TODO validate that one transaction gives the card to the author

    def load_from_data(self, data):
        self.blockId = uuid.UUID(data['blockId'])
        self.previousHash = data['previousHash']
        self.timestamp = datetime.datetime.strptime(
            data['timestamp'],
            "%Y-%m-%d %H:%M:%S.%f%z"
            )
        self.height = data['height']
        self.difficulty = data['difficulty']
        self.nonce = data['nonce']
        self.card = Card(
            cardId = uuid.UUID(data['card']['cardId']),
            pageId = data['card']['pageId']
            )
        self.transactions = [
            Transaction(
                transactionId = uuid.UUID(each['transactionId']),
                timestamp = datetime.datetime.strptime(
                    each['timestamp'],
                    "%Y-%m-%d %H:%M:%S.%f%z"
                    ),
                cardId = uuid.UUID(each['cardId']),
                sender = serialization.load_pem_public_key(each['sender'].encode('utf-8')),
                receiver = serialization.load_pem_public_key(each['receiver'].encode('utf-8')),
                signature = bytes.fromhex(each['signature'])
                ) for each in data['transactions']
            ]
        self.update()

    def as_dict(self):
        return {
            "blockId": str(self.blockId),
            "previousHash": self.previousHash,
            "timestamp": str(self.timestamp),
            "height": self.height,
            "difficulty": self.difficulty,
            "nonce": self.nonce,
            "card": self.card.as_dict(),
            "transactions": [each.as_dict() for each in self.transactions]
            }


    def __str__(self):
        # The hash of the block is the SHA256 hash of what this method returns.
        return json.dumps(self.as_dict(), sort_keys=True)

    class Invalid(Exception):
        def __init__(self, message, statusCode=406):
            super().__init__(message)
            self.message = message
            self.statusCode = statusCode
