import datetime
import io
import json
import uuid

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa

class Transaction():
    def __init__(self, transactionId=None, timestamp=None, cardId=None, sender=None,
                 receiver=None, authorPrivateKey=None, signature=None, data=None):
        self.transactionId = transactionId if transactionId else uuid.uuid4()
        self.timestamp = timestamp if timestamp else datetime.datetime.now(
            datetime.timezone.utc
            )
        self.cardId = cardId
        self.sender = sender
        self.receiver = receiver
        self.signature = signature
        if data:
            self.load_from_data(data)
        if authorPrivateKey:
            self.sign(authorPrivateKey)

    def validate(self):
        # TODO: better error codes
        if not isinstance(self.transactionId, uuid.UUID):
            raise self.Invalid(
                f"Transaction ID should be a UUID not {type(self.transactionId).__name__}."
                )
        if not self.transactionId.version == 4:
            raise self.Invalid(
                f"Transaction ID version should be 4 not {self.transactionId.version}."
                )
        if not isinstance(self.timestamp, datetime.datetime):
            raise self.Invalid(
                f"Timestamp should be a datetime not {type(self.timestamp).__name__}"
                )
        if not self.timestamp.tzinfo == datetime.timezone.utc:
            raise self.Invalid(
                f"Timestamp timezone should be in UTC not {self.timestamp.tzinfo}."
                )
        if not self.timestamp < datetime.datetime.now(datetime.timezone.utc):
            raise self.Invalid(f"Timestamp {self.timestamp} in the future.")
        if not isinstance(self.cardId, uuid.UUID):
            raise self.Invalid(
                f"Card ID should be a UUID not {type(self.cardId).__name__}."
                )
        if not self.cardId.version == 4:
            raise self.Invalid(
                f"Card ID version should be 4 not {self.cardId.version}."
                )
        if not isinstance(self.sender, rsa.RSAPublicKey):
            raise self.Invalid(
                f"Sender should be an RSA public key not {type(self.sender).__name__}."
                )
        if not isinstance(self.receiver, rsa.RSAPublicKey):
            raise self.Invalid(
                f"Receiver should be an RSA public key not {type(self.receiver).__name__}."
                )
        if not isinstance(self.signature, bytes):
            raise self.Invalid(
                f"Signature should be bytes not {type(self.signature).__name__}."
                )
        try:
            self.sender.verify(
                self.signature,
                bytearray(
                    json.dumps({
                        "transactionId": str(self.transactionId),
                        "timestamp": str(self.timestamp),
                        "cardId": str(self.cardId),
                        "sender": self.sender.public_bytes(
                            encoding=serialization.Encoding.PEM,
                            format=serialization.PublicFormat.SubjectPublicKeyInfo
                            ).decode('utf-8'),
                        "receiver": self.receiver.public_bytes(
                            encoding=serialization.Encoding.PEM,
                            format=serialization.PublicFormat.SubjectPublicKeyInfo
                            ).decode('utf-8')
                        }, sort_keys=True).encode('utf-8')
                    ),
                padding.PSS(
                    mgf=padding.MGF1(hashes.SHA256()),
                    salt_length=padding.PSS.MAX_LENGTH
                    ),
                hashes.SHA256()
                )
        # TODO: may need more precision.
        except Exception as e:
            raise self.InvalidSignature(str(e))

    def sign(self, authorPrivateKey):
        self.signature = authorPrivateKey.sign(
            bytearray(
                json.dumps({
                    "transactionId": str(self.transactionId),
                    "timestamp": str(self.timestamp),
                    "cardId": str(self.cardId),
                    "sender": self.sender.public_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PublicFormat.SubjectPublicKeyInfo
                        ).decode('utf-8'),
                    "receiver": self.receiver.public_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PublicFormat.SubjectPublicKeyInfo
                        ).decode('utf-8')
                    }, sort_keys=True).encode('utf-8')
                ),
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH
                ),
            hashes.SHA256()
            )

    def load_from_data(self, data):
        self.transactionId = uuid.UUID(data['transactionId'])
        self.timestamp = datetime.datetime.strptime(
            data['timestamp'],
            "%Y-%m-%d %H:%M:%S.%f%z"
            )
        self.cardId = uuid.UUID(data['cardId'])
        # TODO: why is this a tuple when coming from POST /transactions?
        self.sender = serialization.load_pem_public_key(
            data['sender'].encode('utf-8'),
            backend=default_backend()
            ),
        if isinstance(self.sender, tuple):
            self.sender = self.sender[0]
        self.receiver = serialization.load_pem_public_key(
            data['receiver'].encode('utf-8'),
            backend=default_backend()
            ),
        if isinstance(self.receiver, tuple):
            self.receiver = self.receiver[0]
        self.signature = bytes.fromhex(data['signature'])

    def as_dict(self):
        return {
            "transactionId": str(self.transactionId),
            "timestamp": str(self.timestamp),
            "cardId": str(self.cardId),
            "sender": self.sender.public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo
                ).decode('utf-8'),
            "receiver": self.receiver.public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo
                ).decode('utf-8'),
            "signature": self.signature.hex()
            }

    def __str__(self):
        return json.dumps(self.as_dict(), sort_keys=True)

    class Unauthorized(Exception):
        def __init__(self, message, statusCode=403):
            super().__init__(message)
            self.message = message
            self.statusCode = statusCode

    class Invalid(Exception):
        def __init__(self, message, statusCode=400):
            super().__init__(message)
            self.message = message
            self.statusCode = statusCode

    class AlreadyFulfilled(Exception):
        def __init__(self, message, statusCode=400):
            super().__init__(message)
            self.message = message
            self.statusCode = statusCode

    class InvalidSignature(Exception):
        def __init__(self, message, statusCode=400):
            super().__init__(message)
            self.message = message
            self.statusCode = statusCode
